def save_per_gene_insertions(
    filename,
    tn_coordinates,
    gene_coordinates,
    chr_lengths_cumsum,
    ref_tid_roman,
    aliases_designation,
):
    """This function write in txt file 5 columns of information about transposon insertions separated by tabs. 
    The tabs are: Gene name,Chromosome,Start location,End location,Insertion locations,Reads per insertion location

    Parameters
    ----------
    filename : str
        Path with the filename extension included(e.g "data_file/file.txt") describing where do you want to store the results.
        By default it will be stored in the same location as the bamfile, with the same basename. 
        Example, if the bamfile path is data_file/data_1.bam then the per_gene file will be data_file/data_1.bam_pergene_insertions.txt
    tn_coordinates : dict
         Last output of the function get_insertions _,_, tn_coordinates_per_gene = get_insertions_and_reads(
        gene_coordinates, tncoordinatescopy_array, readnumb_array)
    gene_coordinates : dict 
        Output of the function add_chromosome_length:  gene_coordinates = add_chromosome_length(
        gene_coordinates, chr_lengths_cumsum, ref_tid_roman)
    chr_lengths_cumsum : dict
        last output of the function get_sequence_length : _, chr_lengths_cumsum = get_sequence_length(bam)
    ref_tid_roman : dict 
        Dictionary describing roman names as keys , 
        ref_romannums = chromosomename_roman_to_arabic()[1]
        ref_tid_roman = {key: value for key, value in zip(ref_romannums, ref_tid)}
    aliases_designation : dict
        Last output of the function read_genes  _, _, aliases_designation = read_genes(
        gff_file, essential_file, gene_name_file)
    """

    with open(filename, "w") as f:

        f.write(
            "Gene name\tChromosome\tStart location\tEnd location\tInsertion locations\tReads per insertion location\n"
        )

        for gene in tn_coordinates:
            gene_chrom = ref_tid_roman.get(gene_coordinates.get(gene)[0])
            tncoordinates = [
                ins - chr_lengths_cumsum.get(gene_chrom)
                for ins in tn_coordinates[gene][3]
            ]

            if gene in aliases_designation:
                gene_alias = aliases_designation.get(gene)[0]
            else:
                gene_alias = gene

            f.write(
                gene_alias
                + "\t"
                + str(tn_coordinates[gene][0])
                + "\t"
                + str(tn_coordinates[gene][1] - chr_lengths_cumsum.get(gene_chrom))
                + "\t"
                + str(tn_coordinates[gene][2] - chr_lengths_cumsum.get(gene_chrom))
                + "\t"
                + str(tncoordinates)
                + "\t"
                + str(tn_coordinates[gene][4])
                + "\n"
            )


def save_per_essential_insertions(
    filename,
    tn_coordinates,
    gene_coordinates,
    chr_lengths_cumsum,
    ref_tid_roman,
    aliases_designation,
):
    """This function generates a .txt file with the insertions in the annotated 
    essential genes in WT. 

    Parameters
    ----------
    filename : str
        Path with the filename extension included(e.g "data_file/file.txt") 
        describing where do you want to store the results.
        By default it will be stored in the same location as the bamfile, 
        with the same basename. 
        Example, if the bamfile path is data_file/data_1.bam then the  file will be data_file/data_1.bam_peressential_insertions.txt
    tn_coordinates : dict
         Last output of the function get_insertions _,_, tn_coordinates_per_gene = get_insertions_and_reads(
        gene_coordinates, tncoordinatescopy_array, readnumb_array)
    gene_coordinates : dict 
        Output of the function add_chromosome_length:  gene_coordinates = add_chromosome_length(
        gene_coordinates, chr_lengths_cumsum, ref_tid_roman)
    chr_lengths_cumsum : dict
        last output of the function get_sequence_length : _, chr_lengths_cumsum = get_sequence_length(bam)
    ref_tid_roman : dict 
        Dictionary describing roman names as keys , 
        ref_romannums = chromosomename_roman_to_arabic()[1]
        ref_tid_roman = {key: value for key, value in zip(ref_romannums, ref_tid)}
    aliases_designation : dict
        Last output of the function read_genes  _, _, aliases_designation = read_genes(
        gff_file, essential_file, gene_name_file)
    """

    with open(filename, "w") as f:

        f.write(
            "Essential gene name\tChromosome\tStart location\tEnd location\tInsertion locations\tReads per insertion location\n"
        )

        for gene in tn_coordinates:
            gene_chrom = ref_tid_roman.get(gene_coordinates.get(gene)[0])
            tncoordinates = [
                ins - chr_lengths_cumsum.get(gene_chrom)
                for ins in tn_coordinates[gene][3]
            ]

            if gene in aliases_designation:
                gene_alias = aliases_designation.get(gene)[0]
            else:
                gene_alias = gene

            f.write(
                gene_alias
                + "\t"
                + str(tn_coordinates[gene][0])
                + "\t"
                + str(tn_coordinates[gene][1] - chr_lengths_cumsum.get(gene_chrom))
                + "\t"
                + str(tn_coordinates[gene][2] - chr_lengths_cumsum.get(gene_chrom))
                + "\t"
                + str(tncoordinates)
                + "\t"
                + str(tn_coordinates[gene][4])
                + "\n"
            )



def chromosomename_roman_to_arabic():
    """
    Creates two dictionaries for translating the chromosome names from roman to arabic numerals or vice versa.
    
    Syntax: arabic_to_roman, roman_to_arabic = chromosomename_roman_to_arabic()

    Returns
    -------

    arabic_to_roman : dict
    roman_to_arabic : dict

    """
    num_arabic = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17]
    num_roman = [
        "I",
        "II",
        "III",
        "IV",
        "V",
        "VI",
        "VII",
        "VIII",
        "IX",
        "X",
        "XI",
        "XII",
        "XIII",
        "XIV",
        "XV",
        "XVI",
        "Mito",
    ]

    arabic_to_roman = dict(zip(num_arabic, num_roman))
    roman_to_arabic = dict(zip(num_roman, num_arabic))

    return arabic_to_roman, roman_to_arabic



from .dataframe_from_pergene import dataframe_from_pergenefile
from .volcanoplot import volcano
from .dataframe_from_pergene_helpers import read_pergene_file, reads_per_insertion,essential_genes
from .volcano_helpers import info_from_datasets,make_datafile,apply_stats

__all__ = [
"dataframe_from_pergenefile", "volcano",
"read_pergene_file","reads_per_insertion",
"essential_genes" ,"info_from_datasets",
"make_datafile", "apply_stats"


]


from .chromosome_names_in_files import chromosome_name_bedfile, chromosome_name_wigfile
from .clean_bedwigfiles import cleanfiles
from .essential_genes_names import list_known_essentials
from .transposonread_profileplot_genome import profile_genome
from .profileplot_genome_helpers import (summed_chr,counts_genome,length_genome,middle_chrom_pos,
binned_list)
from .genomicfeatures_dataframe import dna_features
from .dna_features_helpers import (input_region, read_wig_file, read_pergene_file,gene_location,
feature_position,intergenic_regions,checking_features,build_dataframe)

from .read_sgdfeatures import sgd_features


__all__ = [
    "chromosome_name_bedfile",
    "chromosome_name_wigfile",
    "cleanfiles",
    "list_known_essentials",
    "summed_chr","counts_genome",
    "length_genome","middle_chrom_pos",
"binned_list", "profile_genome",
"dna_features", "feature_position", 
"input_region", "read_wig_file" , "read_pergene_file",
"sgd_features", "gene_location" ,"intergenic_regions",
"checking_features","build_dataframe"
]



from transposonmapper.processing.chromosome_names_in_files import chromosome_name_bedfile
from transposonmapper.properties.get_chromosome_position import chromosome_position

import numpy as np 

def summed_chr(chr_length_dict):
    """Create a dictionary where each value is the cumulative sum of all bp in each chromosomes

    Parameters
    ----------
    chr_length_dict : dict
        A dictionary describing the length of each chromosome. 

    Returns
    -------
    dict
        A dictionary where each value corresponds to the cumulative sum of the previous chromosomes lengths. 
    """
    
    chrom_list = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX', 'X', 'XI', 'XII', 'XIII', 'XIV', 'XV', 'XVI']
    summed_chr_length_dict = {}
    summed_chr_length = 0
    for c in chrom_list:
        summed_chr_length_dict[c] = summed_chr_length
        summed_chr_length += chr_length_dict.get(c)    
     
    return summed_chr_length_dict


def length_genome(chr_length_dict):
    
    """Output the length of the genome in bp 

    Parameters
    ----------

    chr_length_dict : dict
        A dictionary describing the length of each chromosome. 

    Returns
    -------
    int
        The length of the genome
    """
    
    chrom_list = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX', 'X', 'XI', 'XII', 'XIII', 'XIV', 'XV', 'XVI']

    l_genome = 0
    for chrom in chrom_list:
        l_genome += int(chr_length_dict.get(chrom))
   
    
    return l_genome



def middle_chrom_pos(chr_length_dict):
    """Defines the middle poit of each chromosome

    Parameters
    ----------
    chr_length_dict : dict
        A dictionary describing the length of each chromosome. 

    Returns
    -------
    list
        A list describing for each chromosome the middle point. 
    """
    
    summed_chr_length_dict=summed_chr(chr_length_dict)
    
    l_genome=length_genome(chr_length_dict)
    
    
    middle_chr_position = []
    c1 = summed_chr_length_dict.get('I')
    for c in summed_chr_length_dict:
        if not c == 'I':
            c2 = summed_chr_length_dict.get(c)
            middle_chr_position.append(c1 + (c2 - c1)/2)
            c1 = c2
            
    c2 = l_genome
    middle_chr_position.append(c1 + (c2 - c1)/2)
    
    return middle_chr_position



def counts_genome(variable,bed_file,gff_file):
    """Counts of reads or the transposons per chromosomes

    Parameters
    ----------
    variable : str
        "transposons" or "reads"
    bed_file : str
        absolute path of the location of the bedfile
    gff_file : str
        absolute path of the location of the gff file 

    Returns
    -------
    numpy.ndarray
        An array of the length of the genome with the counts of each variable per location in the genome. 
    """
    
    with open(bed_file) as f:
        lines = f.readlines()
    
    chrom_names_dict, chrom_start_index_dict, chrom_end_index_dict= chromosome_name_bedfile(bed_file)
    chr_length_dict, chr_start_pos_dict, chr_end_pos_dict = chromosome_position(gff_file)
    
    summed_chr_length_dict=summed_chr(chr_length_dict)
    
    l_genome=length_genome(chr_length_dict)

    allcounts_list = np.zeros(l_genome)
    if variable == "transposons":
        for line in lines[chrom_start_index_dict.get("I"):chrom_end_index_dict.get("XVI")+1]:
            line = line.strip('\n').split()
            chrom_name = [k for k,v in chrom_names_dict.items() if v == line[0].replace("chr",'')][0]
            allcounts_list[summed_chr_length_dict.get(chrom_name) + int(line[1])-1] += 1
    elif variable == "reads":
        for line in lines[chrom_start_index_dict.get("I"):chrom_end_index_dict.get("XVI")+1]:
            line = line.strip('\n').split()
            chrom_name = [k for k,v in chrom_names_dict.items() if v == line[0].replace("chr",'')][0]
            allcounts_list[summed_chr_length_dict.get(chrom_name) + int(line[1])-1] += (int(line[4])-100)/20
    return allcounts_list

def binned_list(allcounts_list,bar_width):
    """A binned list for a histogram of the counts 

    Parameters
    ----------
    allcounts_list : numpy.ndarray
        Output of the counts_genome function
    bar_width : float 
        It could be a function of the length of the genome e.g. bar_width=l_genome/1000

    Returns
    -------
    list
        Binned list 
    """
    
    allcounts_binnedlist = []
    val_counter = 0
    sum_values = 0
    for n in range(len(allcounts_list)):
        if int(val_counter % bar_width) != 0:
            sum_values += allcounts_list[n]
        elif int(val_counter % bar_width) == 0:
            allcounts_binnedlist.append(sum_values)
            sum_values = 0
        val_counter += 1
    allcounts_binnedlist.append(sum_values)
    return allcounts_binnedlist


import matplotlib.pyplot as plt
import numpy as np 
import os 
from scipy import stats

from transposonmapper.statistics import dataframe_from_pergenefile

def make_datafile(path_a,filelist_a,path_b,filelist_b):
    """Assembly the datafile name to analyze 

    Parameters
    ----------
    path_a : str
        Path of the files corresponding to the reference library
    filelist_a : list of str 
        List of the filenames of the different replicates from the reference library.
        It has to have minimum two replicates per library, so the list has to contain
        a minimum of two files. 
    path_b : str 
        Path of the files corresponding to the experimental library
    filelist_b : list of str
        List of the filenames of the different replicates from the experimental library.
        It has to have minimum two replicates per library, so the list has to contain
        a minimum of two files. 


    Returns
    -------
    str
        Complete paths of the reference and the experimental libraries 
    """
    datafiles_list_a = []
    datafiles_list_b = []
    for files in filelist_a:
        datafile = os.path.join(path_a, files)
        assert os.path.isfile(datafile), 'File not found at: %s' % datafile
        datafiles_list_a.append(datafile)
    for files in filelist_b:
        datafile = os.path.join(path_b, files)
        assert os.path.isfile(datafile), 'File not found at: %s' % datafile
        datafiles_list_b.append(datafile)
    return datafiles_list_a,datafiles_list_b


def info_from_datasets(datafiles_list_a,datafiles_list_b,variable,normalize):
    """Read the information contain in the datafiles for the volcano plot 

    Parameters
    ----------
    datafiles_list_a : list of str 
        List of the absolute paths of all the replicates from the 
        reference library. 
    datafiles_list_b : list of str 
        List of the absolute paths of all the replicates from the 
        experimental  library. 
    variable : str
        Magnitude indicating based on what to make the volcano plot. 
        For example: tn_per_gene, read_per_gene or Nreadsperinsrt 

    normalize : bool 
        If True , If set to True, each gene is normalized based on 
        the total count in each dataset (i.e. each file in filelist_)
        

    Returns
    -------
    variable_a_array : numpy.array
    variable_b_array: numpy.array
    volcano_df: pandas.core.frame.DataFrame
    tnread_gene_a: pandas.core.frame.DataFrame
    tnread_gene_b: pandas.core.frame.DataFrame
    """

    tn_per_gene_zeroreplace = 5 #Add 5 insertions to every gene
    read_per_gene_zeroreplace = 25 #Add 25 reads to every gene
    # norm_a = 0
    # norm_b = 0
    for count, datafile_a in enumerate(datafiles_list_a):
        tnread_gene_a = dataframe_from_pergenefile(datafile_a, verbose=False)
        if normalize == True:
            if variable == 'tn_per_gene':
                norm_a = sum(tnread_gene_a.tn_per_gene)#*10**-4
            elif variable == 'read_per_gene':
                norm_a = sum(tnread_gene_a.read_per_gene)#*10**-7
            elif variable == 'Nreadsperinsrt':
                norm_a = sum(tnread_gene_a.Nreadsperinsrt)

        #ADD A CONSTANT TO ALL VALUES TO PREVENT A ZERO DIVISION WHEN DETERMINING THE FOLD CHANGE.
        tnread_gene_a.tn_per_gene = tnread_gene_a.tn_per_gene + tn_per_gene_zeroreplace
        tnread_gene_a.read_per_gene = tnread_gene_a.read_per_gene + read_per_gene_zeroreplace
        tnread_gene_a.Nreadsperinsrt = tnread_gene_a.Nreadsperinsrt + (read_per_gene_zeroreplace/tn_per_gene_zeroreplace)

        if count == 0:
            volcano_df = tnread_gene_a[['gene_names']] #initialize new dataframe with gene_names
            if normalize == True:
                variable_a_array = np.divide(tnread_gene_a[[variable]].to_numpy(), norm_a) #create numpy array to store normalized data
            else:
                variable_a_array = tnread_gene_a[[variable]].to_numpy() #create numpy array to store raw data
        else: 
            if normalize == True:
                variable_a_array = np.append(variable_a_array, np.divide(tnread_gene_a[[variable]].to_numpy(), norm_a), axis=1) #append normalized data
            else:
                variable_a_array = np.append(variable_a_array, tnread_gene_a[[variable]].to_numpy(), axis=1) #append raw data


    for count, datafile_b in enumerate(datafiles_list_b):
        tnread_gene_b = dataframe_from_pergenefile(datafile_b, verbose=False)
        if normalize == True:
            if variable == 'tn_per_gene':
                norm_b = sum(tnread_gene_b.tn_per_gene)#*10**-4
            elif variable == 'read_per_gene':
                norm_b = sum(tnread_gene_b.read_per_gene)#*10**-7
            elif variable == 'Nreadsperinsrt':
                norm_b = sum(tnread_gene_b.Nreadsperinsrt)

        #ADD A CONSTANT TO ALL VALUES TO PREVENT A ZERO DIVISION WHEN DETERMINING THE FOLD CHANGE.
        tnread_gene_b.tn_per_gene = tnread_gene_b.tn_per_gene + tn_per_gene_zeroreplace
        tnread_gene_b.read_per_gene = tnread_gene_b.read_per_gene + read_per_gene_zeroreplace
        tnread_gene_b.Nreadsperinsrt = tnread_gene_b.Nreadsperinsrt + (read_per_gene_zeroreplace/tn_per_gene_zeroreplace)

        if count == 0:
            if normalize == True:
                variable_b_array = np.divide(tnread_gene_b[[variable]].to_numpy(), norm_b)
            else:
                variable_b_array = tnread_gene_b[[variable]].to_numpy()
        else:
            if normalize == True:
                variable_b_array = np.append(variable_b_array, np.divide(tnread_gene_b[[variable]].to_numpy(), norm_b), axis=1)
            else:
                variable_b_array = np.append(variable_b_array, tnread_gene_b[[variable]].to_numpy(), axis=1)

    return variable_a_array,variable_b_array,volcano_df,tnread_gene_a,tnread_gene_b


def apply_stats(variable_a_array,variable_b_array,significance_threshold,volcano_df):
    """This function computes the statistics measure for the volcano plot 

    Parameters
    ----------
    variable_a_array : array
        The values (# of insertions or reads) of the replicates of one library
    variable_b_array : array 
        The values (# of insertions or reads) of the  replicates of the other library
    significance_threshold : float 
        It will use the default value in the volcano function which is 0.01

    Returns
    -------
    dataframe 
        A dataframe containing all the info for the volcano plot. 
    """
    
    ttest_tval_list = [np.nan]*len(variable_a_array) #initialize list for storing t statistics
    ttest_pval_list = [np.nan]*len(variable_a_array) #initialize list for storing p-values
    signif_thres_list = [False]*len(variable_a_array) #initialize boolean list for indicating datapoints with p-value above threshold
    fc_list = [np.nan]*len(variable_a_array) #initialize list for storing fold changes

    for count,val  in enumerate(variable_a_array):

        ttest_val = stats.ttest_ind(variable_a_array[count], variable_b_array[count]) #T-test
        ttest_tval_list[count] = ttest_val[0]
        if not ttest_val[1] == 0: #prevent p=0 to be inputted in log
            ttest_pval_list[count] = -1*np.log10(ttest_val[1])
        else:
            ttest_pval_list[count] = 0
        if ttest_pval_list[count] > -1*np.log10(significance_threshold):
            signif_thres_list[count] = True

    #DETERMINE FOLD CHANGE PER GENE
        if np.mean(variable_b_array[count]) == 0 and np.mean(variable_a_array[count]) == 0:
            fc_list[count] = 0
        else:
            fc_list[count] = np.log2(np.mean(variable_a_array[count]) / np.mean(variable_b_array[count]))

    
    volcano_df['fold_change'] = fc_list
    volcano_df['t_statistic'] = ttest_tval_list
    volcano_df['p_value'] = ttest_pval_list
    volcano_df['significance'] = signif_thres_list

    return volcano_df
 
    




import os, sys
from transposonmapper.properties.get_chromosome_position import chromosome_position
from transposonmapper.processing.chromosome_names_in_files import chromosome_name_bedfile, chromosome_name_wigfile



def cleanfiles(filepath=None, custom_header=None, split_chromosomes=False):
 
    """  This script removes transposon insertions in .bed and .wig files that were mapped outside the chromosomes, creates consistent naming for chromosomes and change the header of files with custom headers.
    This code reads a .bed or .wig file and remove any insertions that were mapped outside a chromosome.
    Mapping of a read outside a chromosome can happen during the alignment and transposon mapping steps and means that the position of an insertions site of a read is larger than the length of the chromosome it is mapped to.
    This function creates a new file with the same name as the inputfile with the extension _clean.bed or _clean.wig.
    This is saved at the same location as the input file.
    In this _clean file the redundant insertions that were mapped outside the chromosome are removed.
    The lengths of the chromosomes are determined the python function 'chromosome_position' which is part of the python module 'chromosome_and_gene_positions.py'.
    This module gets the lengths of the chromosomes from a .gff file downloaded from SGD (https://www.yeastgenome.org/).
    Besides removing the reads outside the chromosomes, it also changes the names of the chromosomes to roman numerals and a custom header can be inputted (optional).
    Finally, the bed and wig files can be split up in separate files for each chromosome. These are placed in _chromosomesplit folder located at the location of the bed or wig file.
    @author: gregoryvanbeek
    Created on Fri Mar  5 15:39:53 2021

    Parameters
    ----------
    filepath : str
        File path of the wig or bed file to analyze
    custom_header : str
        String header to be included in the output file 
    split_chromosomes :  Bool (True/False)
        If true then there will be a folder created for each chromosome , otherwise 
        there will be a file containing all the info for all chromosomes. 
   

    Returns
    -------
    A file with the same basename as the filepath, and in the same location, with the extension : _clean.wig/_clean.bed
        

    """
    
  
    ## checking input files 

    if filepath == None:
        sys.exit(0)
    else:
        assert os.path.isfile(filepath), 'File not found: %s' % filepath

    ## setiing some variables 
    chr_length_dict = chromosome_position(None)[0]

    filepath_splitext = os.path.splitext(filepath)
    exten = filepath_splitext[1]



    num_roman = ['I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI']

    ## main processing 

    if exten == ".bed":
        print("Bed file loaded %s" % filepath)

        chrom_names_dict, chrom_start_line_dict, chrom_end_line_dict = chromosome_name_bedfile(filepath)

        with open(filepath, "r") as f:
            lines = f.readlines()


        with open(filepath_splitext[0]+"_clean.bed", "w") as w:
            #write header
            if custom_header == None or custom_header == "":
                w.write(lines[0])
            else:
                w.write("track name=" + str(custom_header) + " useScore=1\n")

            for chrom in num_roman:
                print("evaluating chromosome %s" % chrom)

                for line in lines[chrom_start_line_dict.get(chrom): chrom_end_line_dict.get(chrom)+1]:
                    line_list = " ".join(line.strip("\n").split()).split(" ")
                    if int(line_list[2]) > chr_length_dict.get(chrom) or int(line_list[1]) < 0:
                        print("Line removed: %s" % line)
                    else:
                        for romanname, chromname in chrom_names_dict.items():
                            if chromname == line_list[0].replace("chr",""):
                                chrom_nameroman = romanname
                        w.write("chr" + str(chrom_nameroman) + " " + str(line_list[1]) + " " + str(line_list[2]) + " " + str(line_list[3]) + " " + str(line_list[4]) + "\n")

            
            for line in lines[chrom_end_line_dict.get("XVI")+1:]:
                line_list = " ".join(line.strip("\n").split()).split(" ")
                w.write("chrM" + " " + str(line_list[1]) + " " + str(line_list[2]) + " " + str(line_list[3]) + " " + str(line_list[4]) + "\n")


        if split_chromosomes == True:
            path = os.path.dirname(filepath)
            name = os.path.splitext(os.path.basename(filepath_splitext[0]+"_clean.bed"))[0]

            directoryname = os.path.join(path, name + '_chromosomesplit')

            if not os.path.exists(directoryname):
                os.mkdir(directoryname)

            chromosome_names = ['I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI']

            chrom_names_dict, chrom_start_line_dict, chrom_end_line_dict = chromosome_name_bedfile(os.path.join(path, name+".bed"))

            with open(os.path.join(path,name+".bed"), 'r') as f:
                lines = f.readlines()
            header = lines[0]


            for chrom in chromosome_names:
                outputfile = os.path.join(directoryname, name + '_' + str(chrom) + '.bed')
                with open(outputfile, 'w+') as f:
                    f.write(header)
                    for l in range(chrom_start_line_dict.get(chrom), chrom_end_line_dict.get(chrom)+1):
                        f.write(lines[l])

                outputfile = os.path.join(directoryname, name + '_M.bed')
                with open(outputfile, 'w+') as f:
                    f.write(header)
                    for l in range(chrom_end_line_dict.get(chromosome_names[-1])+1, len(lines)):
                        f.write(lines[l])




    elif exten == ".wig":
        print("Wig file loaded %s" % filepath)

        chrom_names_dict, chrom_start_line_dict, chrom_end_line_dict = chromosome_name_wigfile(filepath)

        with open(filepath, 'r') as f:
            lines = f.readlines()

        with open(filepath_splitext[0]+"_clean.wig", "w") as w:
            #write header
            if custom_header == None or custom_header == "":
                w.write(lines[0].replace(',',''))
            else:
                w.write("track type=wiggle_0 maxheightPixels=60 name=" + str(custom_header) + "\n")

            for chrom in num_roman:
                print("evaluating chromosome %s" % chrom)

                #replace chromosome names from reference genome with roman numerals
                chrom_headerline = lines[chrom_start_line_dict.get(chrom) - 1]
                chrom_nameline = chrom_headerline.split("=")[1].strip("\n").replace("chr","")
                for romanname, chromname in chrom_names_dict.items():
                    if chromname.replace("chr","") == chrom_nameline:
                        chrom_nameroman = romanname
                w.write("variablestep chrom=chr" + str(chrom_nameroman) + "\n") #write header for each chromosome
                for line in lines[chrom_start_line_dict.get(chrom): chrom_end_line_dict.get(chrom)]: #no '+1' in for loop, this is only for bed file
                    line_list = " ".join(line.strip("\n").split()).split(" ")
                    if int(line_list[0]) > chr_length_dict.get(chrom) or int(line_list[0]) < 0:
                        print("Line removed: %s" % line)
                    else:
                        w.write(line)


            w.write("variablestep chrom=chrM\n")
            for line in lines[chrom_end_line_dict.get("XVI")+1:]:
                w.write(line)



        if split_chromosomes == True:
            path = os.path.dirname(filepath)
            name = os.path.splitext(os.path.basename(filepath_splitext[0]+"_clean.wig"))[0]

            directoryname = os.path.join(path, name + '_chromosomesplit')

            if not os.path.exists(directoryname):
                os.mkdir(directoryname)

            chromosome_names = ['I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','M']

            chrom_names_dict, chrom_start_line_dict, chrom_end_line_dict = chromosome_name_wigfile(os.path.join(path, name+".wig"))

            with open(os.path.join(path,name+".wig"), 'r') as f:
                lines = f.readlines()
            header = lines[0]


            for chrom in chromosome_names:
                outputfile = os.path.join(directoryname, name + '_' + str(chrom) + '.wig')
                with open(outputfile, 'w+') as f:
                    f.write(header)
                    for l in range(chrom_start_line_dict.get(chrom)-1, chrom_end_line_dict.get(chrom)):
                        f.write(lines[l])

    else:
        print("Extension not recognized")





import os

from .get_gene_position import gene_position
from .gene_aliases import gene_aliases


def get_coordinates_genes(path: str = "", data_files: dict = {}):
    """
    Get coordinates of all genes

    Parameters
    ----------
    path : str
    data_files : dict

    Returns
    -------
    essential_coordinates : dict    
    aliases_designation : dict

    """

    essential_coordinates = {}

    # Get position genes
    if "gff3" in data_files:
        file_path = os.path.join(path, data_files["gff3"])
        gene_coordinates = gene_position(file_path)
    else:
        raise ValueError("gff3 type not found in data")

    # Get all annotated essential genes
    if "essential_genes" in data_files:
        file_path = os.path.join(path, data_files["essentials"])
        with open(file_path, "r") as f:
            genes = f.readlines()[1:]
            for gene in genes:
                name = gene.strip("\n")
                essential_coordinates[name] = gene_coordinates.get(name).copy()
    else:
        raise ValueError("essentials not found in data")

    # Get aliases of all genes
    if "gene_names" in data_files:
        file_path = os.path.join(path, "Yeast_Protein_Names.txt")
        aliases_designation = gene_aliases(file_path)[0]  #'YMR056C' \ ['AAC1'], ...
    else:
        raise ValueError("gene_names not found in data")

    return essential_coordinates, aliases_designation



from transposonmapper.properties.get_gene_position import gene_position
from transposonmapper.properties.gene_aliases import gene_aliases


def read_genes(gff_file, essentials_file, gene_names_file):
    
    """  This function reads the useful information inside the gff_file, essentials_file and
    gene_names_file. 
    For the gff_file and essentials_file extracts the gene coordinates , specifying the chromosome, start ,end 
    and direction. 
    For the gene_names_files it translates the systematic name into the standard name.
    

    Parameters
    ----------
    gff_file : .gff3
        Annotated genome from Saccharomyces cerevisiae (baker's yeast) (https://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/146/045/GCF_000146045.2_R64/GCF_000146045.2_R64_genomic.gff.gz)
    essentials_file : .txt
        Essentials genes annotated from yeast , written using the systematic name standard , all in one column
    gene_names_file : .txt
        This documents lists all the Saccharomyces cerevisiae S288c entries present in
        this release of UniProtKB/Swiss-Prot. Yeast (Saccharomyces cerevisiae): entries, 
        gene names and cross-references to SGD. 
        Release:     2021_01 of 10-Feb-2021

    Returns
    -------
    dict:
        gene_coordinates : a dict specifying for each gene the chromosome number the gene
        belongs to, the start gene coordinate, the end gene coordinate and the strand direction ('+' or '-'). 
    dict:
        essential_coordinates: a dict specifying for each annotated essential gene the chromosome number the gene
        belongs to, the start gene coordinate, the end gene coordinate and the strand direction ('+' or '-'). 
    dict:
        aliases_designation: a dict that for each systematic gene name specify the standard gene name.


    """

    # Get gene position
    gene_coordinates = gene_position(gff_file)  #'YAL069W' | ['I', 335, 649], ...

    # Get all annotated essential genes
    essential_coordinates = {}
    with open(essentials_file, "r") as f:
        genes = f.readlines()[1:]
        for gene in genes:
            name = gene.strip("\n")
            essential_coordinates[name] = gene_coordinates.get(name).copy()

    # Get aliases of all genes
    aliases_designation = gene_aliases(gene_names_file)[0]  #'YMR056C' \ ['AAC1'], ...

    return gene_coordinates, essential_coordinates, aliases_designation



# =============================================================================
# To import the functions, type:
#    import sys
#    sys.path.insert(1,r'full\path\to\this\file')
#    from essential_genes import list_known_essentials
# =============================================================================

import os
from transposonmapper.importing import load_default_files

def list_known_essentials(input_files = None, headerlines=3, verbose=True):
    """Get all known essential genes from two different files and combine them in one list.
        Input is a list of of paths where files can be found with the known essential genes.
        A default list is implemented using two files present in the same folder as this file.
        It is expected that the files contain genes in a single column and nothing else.
        An option can be set for the number of headerlines, which by default is set to 3.
        The output is a list containing all the genes present in all files given in the input.
        
        Note when using the default files: The length of the output list exceed the number of known essential genes as the list sometimes contains both the standard name and the systematic name of a gene.

    Parameters
    ----------
    input_files : str, optional
        File path for the essential file in your file system, by default None
    headerlines : int, optional
         by default 3
    verbose : bool, optional
        To show explanations if True, by default True
    """
    
    if input_files == None:
        
        # If necessary, load default files
        gff_file, essential_genes_files, gene_name_file = load_default_files(
            gff_file=None, essentials_file=None, gene_names_file=None
        )

        # Verify presence of files
        data_files = {
            "gff3": gff_file,
            "essentials": essential_genes_files,
            "gene_names": gene_name_file,
        }

        for filetype, file_path in data_files.items():
            assert file_path, f"{filetype} not found at {file_path}"
        
        input_files=essential_genes_files

    else:
        essential_genes_files = input_files


    known_essential_gene_list = []
    #making sure essential_genes_files is a list
    if type(essential_genes_files)!=list:
        essential_genes_files=[input_files]
        
    for files in essential_genes_files:
        if verbose == True:
            print('Reading file :',files)
        with open(files) as f:
            for header_lines in range(headerlines):
                next(f)
            for lines in f:
                known_essential_gene_list.append(lines.rstrip('\n'))
                
    
    return(known_essential_gene_list)
# #%%
# if __name__ == '__main__':
#     known_essential_gene_list = list_known_essentials()





import os, sys
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np 
import pkg_resources


from transposonmapper.properties import chromosome_position,gene_aliases
from transposonmapper.processing import chromosome_name_wigfile

from transposonmapper.processing.read_sgdfeatures import sgd_features


from transposonmapper.importing import (
    load_default_files, load_sgd_tab
)


from transposonmapper.processing.dna_features_helpers import (build_dataframe, input_region, intergenic_regions, read_pergene_file, 
                                                              read_wig_file,gene_location,checking_features)
from transposonmapper.plotting.dna_features_plot import dna_features_plot

def dna_features(region, wig_file, pergene_insertions_file, variable="reads", plotting=True, savefigure=False, verbose=True):
    """This scripts takes a user defined genomic region (i.e. chromosome number, region or gene) and creates a dataframe including information about all genomic features in the chromosome (i.e. genes, nc-DNA etc.).
    This can be used to determine the number of reads outside the genes to use this for normalization of the number of reads in the genes.
    Output is a dataframe including major information about all genomic features and optionally a barplot indicating the number of transposons per genomic region.
    A genomic region is here defined as a gene (separated as annotated essential and not essential), telomere, centromere, ars etc.
    This can be used for identifying neutral regions (i.e. genomic regions that, if inhibited, do not influence the fitness of the cells).
    This function can be used for normalizing the transposon insertions per gene using the neutral regions.
    
              
    
    Parameters
    ----------
    region : str
        - Region: e.g. chromosome number (either a normal number between 1 and 16 or in roman numerals between I and XVI), a list like ['V', 0, 14790] which creates a barplot between basepair 0 and 14790) or a genename.

    wig_file : str
        absolute path for the wig file location
    pergene_insertions_file : str 
        asbsoulte path for the _pergene_insertions.txt file location 
    variable : str, optional
        By default "reads". It could be "transposons"or "reads". This would be used for the plotting if True 
    plotting : bool, optional
        Whether or not producing a bar plot with the reads/insertions per genomic location in the region, by default True
    savefigure : bool, optional
        Whether or not saving the plot in the same folder as the datafiles, by default False
    verbose : bool, optional
        Determines how much textual feedback is given. When set to False, only warnings will be shown. By default True

    Returns
    -------
    dataframe
         Dataframe containing information about the selected chromosome. 

    
    """

    # If necessary, load default files
    gff_file, essentials_file, gene_information_file = load_default_files()
    sgd_features_file=load_sgd_tab()

    # Verify presence of files
    data_files = {
        "gff3": gff_file,
        "essentials": essentials_file,
        "gene_names": gene_information_file,
        "sgd_features": sgd_features_file
    }

    for filetype, file_path in data_files.items():
        assert file_path, f"{filetype} not found at {file_path}"


    variable = variable.lower()
    if plotting == True:
        create_plottitle = ''

# DETERMINE INPUTTED REGION

    roi_start,roi_end,region_type,chrom=input_region(region=region,verbose=verbose)

    

#READ WIG FILE FOR GETTING LOCATIONS OF ALL TN INSERTIONS

    insrt_in_chrom_list,reads_in_chrom_list=read_wig_file(wig_file=wig_file,chrom=chrom)


# READ PERGENE_INSERTIONS FILE FOR LOCATION OF ALL INSERTIONS PER EACH GENE.

    gene_position_dict=read_pergene_file(pergene_insertions_file=pergene_insertions_file,chrom=chrom)

# DETERMINE THE LOCATION GENOMIC FEATURES IN THE CURRENT CHROMOSOME AND STORE THIS IN A DICTIONARY

    dna_dict,start_chr,end_chr,len_chr,feature_orf_dict=gene_location(chrom,gene_position_dict,verbose)

## GET FEATURES FROM INTERGENIC REGIONS 

    dna_dict_new,genomicregions_list=intergenic_regions(chrom,start_chr,dna_dict)


    ### TEST IF ELEMENTS IN FEATURE_ORF_DICT FOR SELECTED CHROMOSOME ARE THE SAME AS THE GENES IN GENE_POSITION_DICT BY CREATING THE DICTIONARY FEATURE_POSITION_DICT CONTAINING ALL THE GENES IN FEATURE_ORF_DICT WITH THEIR CORRESPONDING POSITION IN THE CHROMOSOME
    checking_features(feature_orf_dict,chrom,gene_position_dict,verbose)

    dna_df2=build_dataframe(dna_dict_new,start_chr,end_chr,insrt_in_chrom_list,reads_in_chrom_list,genomicregions_list,chrom)

    #PRINT INFORMATION FOR THE SELECTED GENE
    if region_type == 'Gene':
        for region_info in dna_df2.itertuples():
            if region_info.Feature_name == region.upper() or region_info.Standard_name == region.upper():
                print(region_info)

    
    # CREATE BAR PLOT 
    dna_features_plot(chrom, dna_df2,roi_start,roi_end,len_chr,plotting,variable)

    if savefigure == True:
        file_dirname=pkg_resources.resource_filename("transposonmapper", "data_files/")
        if variable == 'reads':
            saving_name = os.path.join(file_dirname,'GenomicFeaturesReads_Barplot_Chrom'+chrom+'_NonNormalized')
        else:
            saving_name = os.path.join(file_dirname,'GenomicFeaturesInsertions_Barplot_Chrom'+chrom+'_NonNormalized')
        plt.savefig(saving_name, orientation='landscape', dpi=200)
        plt.close()

    return(dna_df2)









import os
import numpy as np
import re 

from transposonmapper.importing import load_default_files
from transposonmapper.processing import list_known_essentials

def read_pergene_file(pergenefile):
    """It reads the content of the pergene file , one of the outputs of Transposonmapper 

    Parameters
    ----------
    pergenefile : str
        absolute path to the pergene.txt file , one of the outputs of the transposonmapper module

    Returns
    -------
    list
        Gene names list
    list
        Insertion list
    list 
        Reads list 
    """

    assert os.path.isfile(pergenefile), 'File not found at: %s' % pergenefile

    with open(pergenefile) as f:
        lines = f.readlines()[1:] #skip header

    genenames_list = [np.nan]*len(lines)
    tnpergene_list = [np.nan]*len(lines)
    readpergene_list = [np.nan]*len(lines) 

    line_counter = 0
    for line in lines:
        line_split = re.split(' |\t', line.strip('\n'))
        l = [x for x in line_split if x]

        
        genenames_list[line_counter] = l[0]
        tnpergene_list[line_counter] = int(l[1])
        readpergene_list[line_counter] = int(l[2])

        line_counter += 1
    return genenames_list,tnpergene_list,readpergene_list,lines


def reads_per_insertion(tnpergene_list,readpergene_list,lines):
    """It computes the reads per insertion following the formula:
    reads/(insertions-1) if the number of insertions is higher than 5, 
    if not then the reads per insertion will be 0. 

    Parameters
    ----------
    tnpergene_list : list
        A list with all insertions
    readpergene_list : list 
        A list of the reads 
    lines : int
        Number of genes mapped to in the reference genome 

    Returns
    -------
    list
         A list containing all the reads per insertions per gene. 
    """

    readperinspergene_list = [np.nan]*len(lines)
    for i in range(len(tnpergene_list)):
        if not tnpergene_list[i] < 5:
            readperinspergene_list[i] = readpergene_list[i] / (tnpergene_list[i] -1)
        else:
            readperinspergene_list[i] = 0

    return readperinspergene_list
    
def essential_genes(genenames_list,lines):
    """It provides a list of essential genes 

    Parameters
    ----------
    genenames_list : list
        A list will al genes names that were mapped to the reference genome
    lines : int 
        Number of genes in total 

    Returns
    -------
    list 
        List of essential genes 
    """

    _,essential_genes_list,_=load_default_files()
    known_essential_gene_list = list_known_essentials(essential_genes_list)
    geneessentiality_list = [None]*len(lines)
    for i in range(len(genenames_list)):
        if genenames_list[i] in known_essential_gene_list:
            geneessentiality_list[i] = True
        else:
            geneessentiality_list[i] = False
        
    return geneessentiality_list


import numpy as np

def get_insertions_and_reads(coordinates, tn_coordinates, readnumb_array):
    """This function computes the total number of transposons per gene , the number of reads per gene and
    the distribution  of transposons along the gene.

    Parameters
    ----------
    coordinates : dict 
        This is the output of the function add_chromosome_length(coordinates, chr_lengths_cumsum, ref_tid_roman)
    tn_coordinates : numpy.array
        This is the output of the function add_chromosome_length_inserts(coordinates, ref_names, chr_lengths)
    readnumb_array : numpy.array
        This is the 1st output of the function get_reads(bam)

    Returns
    -------
     dict 
        A dict which every key corresponds with each gene and each value with the total number of transposons found in that gene
    
     dict
        A dict which every key corresponds with each gene and each value with the total number of reads for all the transposons found in that gene

     dict 
        A dict which every key corresponds with each gene and each value with a  list of 4 elements:
            - the chromosome number
            - gene start position
            - gene end position
            - distribution of reads per transposon found inside the gene 

    """

    tn_per_gene = {}
    reads_per_gene = {}
    tn_coordinates_per_gene = {}

    for gene in coordinates:
        xx = np.where(np.logical_and(tn_coordinates[:,1] >= coordinates.get(gene)[1], tn_coordinates[:,1] <= coordinates.get(gene)[2])) #get all insertions within range of current gene
        tn_per_gene[gene] = np.size(xx)
        reads_per_gene[gene] = sum(readnumb_array[xx]) - max(readnumb_array[xx], default=0) #REMOVE LARGEST VALUE TO REDUCE NOISE

        if np.size(xx) > 0:
            tn_coordinates_per_gene[gene] = [coordinates.get(gene)[0], coordinates.get(gene)[1], coordinates.get(gene)[2], list(tn_coordinates[xx[0][0]:xx[0][-1]+1, 1]), list(readnumb_array[xx])]
        else:
            tn_coordinates_per_gene[gene] = [coordinates.get(gene)[0], coordinates.get(gene)[1], coordinates.get(gene)[2], [], []]

    return tn_per_gene, reads_per_gene, tn_coordinates_per_gene


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This is a tool developed for analysing transposon insertions for experiments using SAturated Transposon Analysis in Yeast (SATAY).
This python code contains one function called transposonmapper().
For more information about this code and the project, see https://satay-ll.github.io/SATAY-jupyter-book/Introduction.html

This code is based on the Matlab code created by the Kornmann lab which is available at: sites.google.com/site/satayusers/

__Author__ = Gregory van Beek. LaanLab, department of Bionanoscience, Delft University of Technology
__version__ = 1.5
__Date last update__ = 2021-01-11

Version history:
    1.1; Added code for creating two text files for storing insertion locations per gene and per essential gene [2020-07-27]
    1.2; Improved searching algorithm for essential genes [2020-08-06]
    1.3; Load file containing all essential genes so that a search for essential genes in multiple file is not needed anymore. This file is created using Create_EssentialGenes_list.py located in the same directory as this code [2020-08-07]
    1.4; Fixed bug where the gene position and transposon insertion location did not start at zero for each chromosome, causing confusing values to be stored in the _pergene_insertions.txt and _peressential_insertions.txt files [2020-08-09]
    1.5; Added functionality to handle all possible sam flags in the alignment file (bam-file) instead of only flag=0 or flag=16. This is needed for the function to handle paired-end sequencing data [2021-01-11]
"""



# Local imports
from transposonmapper.properties import (
    get_chromosome_names,
    get_sequence_length,
)

from transposonmapper.mapping import (
    get_reads,
    add_chromosome_length,
    add_chromosome_length_inserts,
    get_insertions_and_reads,
)

from transposonmapper.utils import chromosomename_roman_to_arabic

from transposonmapper.importing import (
    load_default_files,
    read_genes,
)

from transposonmapper.exporting import (
    save_as_bed,
    save_per_gene,
    save_per_gene_insertions,
    save_per_essential_insertions,
    save_as_wig
)

import sys

def transposonmapper(bamfile, gff_file=None, essential_file=None, gene_name_file=None):
    """This function is created for analysis of SATAY data using the species Saccharomyces Cerevisiae.
    The function assumes that the reads are already aligned to a reference genome.
    The input data should be a .bam-file and the location where the .bam-file is stored should also contain an index file (.bam.bai-file, which for example can be created using sambamba).
    The function uses the pysam package for handling bam files (see pysam.readthedocs.io/en/latest/index.html) and therefore this function only runs on Linux systems with SAMTools installed.

   
    
    Parameters
    ----------
    bamfile : str, required
        Path to the bamfile. This location should also contain the .bam.bai index file (does not need to be input in this function).
    gff_file : str, optional
        Path to a .gff-file including all gene information (e.g. downloaded from SGD). 
        Default file is 'Saccharomyces_cerevisiae.R64-1-1.99.gff3'., by default None
    essential_file : str, optional
        Path to a .txt file containing a list all essential genes. Every line should consist of a single essential gene and the file should have one header line. 
        Ideally this file is created using 'Create_EssentialGenes_list.py'. Default file is 'Cerevisiae_AllEssentialGenes_List.txt'., by default None
    gene_name_file : str, optional
        Path to text file that includes aliases for all genes. Default file is 'Yeast_Protein_Names.txt', by default None

   Returns
   -------
    A set of files
         It outputs the following files that store information regarding the location of all insertions:
        - .bed-file: Includes all individual basepair locations of the whole genome where at least one transposon has been mapped and the number of insertions for each locations (the number of reads) according to the Browser Extensible Data (bed) format.
                    A distinction is made between reads that had a different reading orientation during sequencing. The number of reads are stored using the equation #reads*20+100 (e.g. 2 reads is stored as 140).
        - .wig-file: Includes all individual basepair locations of the whole genome where at least one transposon has been mapped and the number of insertions for each locations (the number of reads) according to the Wiggle (wig) format.
                    In this file no distinction is made between reads that had a different reading orientation during sequencing. The number of reads are stored as the absolute count.
        - _pergene.txt-file: Includes all genes (currently 6600) with the total number of insertions and number of reads within the genomic region of the gene.
        - _peressential.txt-file: Includes all annotated essential genes (currently 1186) with the total number of insertions and number of reads within the genomic region of the gene.
        - _pergene_insertions.txt-file: Includes all genes with their genomic location (i.e. chromosome number, start and end position) and the locations of all insertions within the gene location. It also include the number number of reads per insertions.
        - _peressential_insertions.txt-file: Includes all essential genes with their genomic location (i.e. chromosome number, start and end position) and the locations of all insertions within the gene location. It also include the number number of reads per insertions.
          (note that in the latter two files, the genomic locations are continous, for example chromosome II does not start at 0, but at 'length chromosome I + 1' etc.).
    The output files are saved at the location of the input file using the same name as the input file, but with the corresponding extension.
    
    
    """

    # If necessary, load default files
    gff_file, essential_file, gene_name_file = load_default_files(
        gff_file, essential_file, gene_name_file
    )

    # Verify presence of files
    data_files = {
        "bam": bamfile,
        "gff3": gff_file,
        "essentials": essential_file,
        "gene_names": gene_name_file,
    }

    for filetype, file_path in data_files.items():
        assert file_path, f"{filetype} not found at {file_path}"

    # Read files for all genes and all essential genes
    print("Getting coordinates of all genes ...")
    gene_coordinates, essential_coordinates, aliases_designation = read_genes(
        gff_file, essential_file, gene_name_file
    )

    try:
        import pysam
    except ImportError:
        print("Failed to import pysam")
        sys.exit(1)
    # Read bam file
    bam = pysam.AlignmentFile(bamfile, "rb")

    # Get names of all chromosomes as stored in the bam file
    ref_tid = get_chromosome_names(bam)
    ref_names = list(ref_tid.keys())

    # Convert chromosome names in data file to roman numerals
    ref_romannums = chromosomename_roman_to_arabic()[1]
    ref_tid_roman = {key: value for key, value in zip(ref_romannums, ref_tid)}

    # Get sequence lengths of all chromosomes
    chr_lengths, chr_lengths_cumsum = get_sequence_length(bam)

    # Get all reads within a specified genomic region
    readnumb_array, tncoordinates_array, tncoordinatescopy_array = get_reads(bam)

    #%% CONCATENATE ALL CHROMOSOMES

    # For each insertion location, add the length of all previous chromosomes
    tncoordinatescopy_array = add_chromosome_length_inserts(
        tncoordinatescopy_array, ref_names, chr_lengths
    )

    # For each gene location, add the length of all previous chromosomes
    gene_coordinates = add_chromosome_length(
        gene_coordinates, chr_lengths_cumsum, ref_tid_roman
    )

    # For each essential gene location, add the length of all previous chromosomes
    essential_coordinates = add_chromosome_length(
        essential_coordinates, chr_lengths_cumsum, ref_tid_roman
    )

    # GET NUMBER OF TRANSPOSONS AND READS PER GENE
    print("Get number of insertions and reads per gene ...")

    # All genes
    tn_per_gene, reads_per_gene, tn_coordinates_per_gene = get_insertions_and_reads(
        gene_coordinates, tncoordinatescopy_array, readnumb_array
    )

    # Only essential genes
    (
        tn_per_essential,
        reads_per_essential,
        tn_coordinates_per_essential,
    ) = get_insertions_and_reads(
        essential_coordinates, tncoordinatescopy_array, readnumb_array
    )

    # CREATE BED FILE
    bedfile = bamfile + ".bed"
    print("Writing bed file at: ", bedfile)
    print("")
    save_as_bed(bedfile, tncoordinates_array, ref_tid, readnumb_array)

    # CREATE TEXT FILE WITH TRANSPOSONS AND READS PER GENE
    # NOTE THAT THE TRANSPOSON WITH THE HIGHEST READ COUNT IS IGNORED.
    # E.G. IF THIS FILE IS COMPARED WITH THE _PERGENE_INSERTIONS.TXT FILE THE READS DON'T ADD UP (SEE https://groups.google.com/forum/#!category-topic/satayusers/bioinformatics/uaTpKsmgU6Q)
    # TOO REMOVE THIS HACK, CHANGE THE INITIALIZATION OF THE VARIABLE readpergene
    per_gene_file = bamfile + "_pergene.txt"
    print("Writing pergene.txt file at: ", per_gene_file)
    print("")

    save_per_gene(per_gene_file, tn_per_gene, reads_per_gene, aliases_designation)

    # CREATE TEXT FILE TRANSPOSONS AND READS PER ESSENTIAL GENE
    per_essential_file = bamfile + "_peressential.txt"
    print("Writing peressential.txt file at: ", per_essential_file)
    print("")
    save_per_gene(
        per_essential_file, tn_per_essential, reads_per_essential, aliases_designation
    )

    # CREATE TEXT FILE WITH LOCATION OF INSERTIONS AND READS PER GENE
    per_gene_insertions_file = bamfile + "_pergene_insertions.txt"
    print("Witing pergene_insertions.txt file at: ", per_gene_insertions_file)
    print("")

    save_per_gene_insertions(
        per_gene_insertions_file,
        tn_coordinates_per_gene,
        gene_coordinates,
        chr_lengths_cumsum,
        ref_tid_roman,
        aliases_designation,
    )

    # CREATE TEXT FILE WITH LOCATION OF INSERTIONS AND READS PER ESSENTIAL GENE
    per_essential_insertions_file = bamfile + "_peressential_insertions.txt"
    print(
        "Writing peressential_insertions.txt file at: ", per_essential_insertions_file
    )
    print("")

    save_per_essential_insertions(
        per_essential_insertions_file,
        tn_coordinates_per_essential,
        gene_coordinates,
        chr_lengths_cumsum,
        ref_tid_roman,
        aliases_designation,
    )

    # ADD INSERTIONS AT SAME LOCATION BUT WITH DIFFERENT ORIENTATIONS TOGETHER (FOR STORING IN WIG-FILE)
    wigfile = bamfile + ".wig"
    print("Writing wig file at: ", wigfile)
    print("")

    save_as_wig(wigfile, tncoordinates_array, ref_tid, readnumb_array)


#%%
if __name__ == "__main__":
    bamfile = "transposonmapper/data_files/files4test/SRR062634.filt_trimmed.sorted.bam"
    transposonmapper(bamfile=bamfile)




# -*- coding: utf-8 -*-
"""
Created on Mon Jan  4 09:38:16 2021

@author: gregoryvanbeek
"""

#%%INPUT
flag = 1040
verbose=True

#%%

def samflags(flag=0, verbose=True):
    """This script converts a decimal flag to binary and get the corresponding properties according to the sam-flag standard.
    The code is based on the explanation given here https://davetang.org/muse/2014/03/06/understanding-bam-flags/
    For manual checking sam flags, check https://broadinstitute.github.io/picard/explain-flags.html
    The input is a decimal number.

    Parameters
    ----------
    flag : int, optional
         by default 0
    verbose : bool, optional
         by default True
    """

    flag_binary = format(flag, '012b') # '#012b' to start the string with '0b'. 12 indicated that the string has length 12.


    prop_dict = {1: 'read paired',
                 2: 'read mapped in proper pair',
                 3: 'read unmapped',
                 4: 'mate unmapped',
                 5: 'read reverse strand',
                 6: 'mate reverse strand',
                 7: 'first in pair',
                 8: 'second in pair',
                 9: 'not primary alignment',
                 10: 'read fails platform/vendor quality checks',
                 11: 'read is PCR or optical duplicate',
                 12: 'supplementary alignment'}


    counter = 1
    flagprop_list = []
    for b in reversed(flag_binary):
        if int(b) == 1:
            flagprop_list.append(prop_dict.get(counter))
        counter += 1

    if verbose == True:
        print('Entered decimal flag = %i' % flag)
        print('Corresponding binary flag = %s' % flag_binary)
        print(flagprop_list)
        print('')


    return(flag_binary, flagprop_list)



#%%
if __name__ == '__main__':
    flag_binary, flagproperties = samflags(flag=flag, verbose=verbose)



import os
import pkg_resources

def load_default_files(gff_file=None, essentials_file=None, gene_names_file=None):
    """This function loads some files that have a recurrent use throughout the pipeline.
    It will look inside the satay/data_files folder for the files if the input is None. 
    Otherwise it will return the same input file.  
    


    Parameters
    ----------
    gff_file : .gff3, optional
         Annotated genome from Saccharomyces cerevisiae (baker's yeast) 
        (https://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/146/045/GCF_000146045.2_R64/GCF_000146045.2_R64_genomic.gff.gz), by default None
    essentials_file : .txt, optional
        Essentials genes annotated from yeast , written using the systematic name standard , 
        all in one column, by default None
    gene_names_file : .txt, optional
        This documents lists all the Saccharomyces cerevisiae S288c entries present in
        this release of UniProtKB/Swiss-Prot. Yeast (Saccharomyces cerevisiae): entries, 
        gene names and cross-references to SGD. 
        Release:     2021_01 of 10-Feb-2021, by default None

    Returns
    -------
    str: 

        The correct path of the .gff3 input required file for the further analysis. 
    str: 

        The correct path of the essential file input required file for the further analysis. 
    str: 

        The correct path of the gene_name_file required file for the further analysis.
    """

    default_path = pkg_resources.resource_filename("transposonmapper", "data_files/")
    if gff_file is None:
        gff_file = os.path.join(
            default_path, "Saccharomyces_cerevisiae.R64-1-1.99.gff3"
        )
    if essentials_file is None:
        essentials_file = os.path.join(
            default_path, "Cerevisiae_AllEssentialGenes_List.txt"
        )
    if gene_names_file is None:
        gene_names_file = os.path.join(default_path, "Yeast_Protein_Names.txt")


    return gff_file, essentials_file, gene_names_file


def load_sgd_tab(sgd_features_file=None):
    """This function loads the file SGD_features.tab
    The latest version of the SGD_features.tab file is based on Genome Version R64-2-1.
    If a specific file is provided it will output that file , otherwise , if it is set to None then 
    it will give the standard file provided in the package. 
        
   

    Parameters
    ----------
    sgd_features_file : str, optional
        The latest version of the SGD_features.tab file is based on Genome Version R64-2-1., by default None

    Returns
    -------
    str: 
        The path corresponding to this file in the package 
    """
    default_path = pkg_resources.resource_filename("transposonmapper", "data_files/")
    if sgd_features_file is None: 

        sgd_features_file = os.path.join(default_path,'SGD_features.tab')

    return sgd_features_file




import numpy as np

def add_chromosome_length_inserts(coordinates, ref_names, chr_lengths):
    """ For each insertion location, add the length of all previous chromosomes

    Parameters
    ----------
    numpy.array
        Third output from get_reads(bam)
    list
    Output from the following : 
        ref_tid = get_chromosome_names(bam)

        ref_names = list(ref_tid.keys())

    dict
        First output from get_sequence_length(bam)
   

    Returns
    -------
    numpy.array
      For each insertion location, add the length of all previous chromosomes 
        
    """
    
    ll = 0
    for ii in range(1,len(ref_names)):
        ll += chr_lengths[ref_names[ii-1]]
        aa = np.where(coordinates[:,0] == ii + 1)
        coordinates[aa,1] = coordinates[aa,1] + ll

    return coordinates


def add_chromosome_length(coordinates, chr_lengths_cumsum, ref_tid_roman):
    
    """This function returns a dictionary that for every gene , there is the chromosome number
    information of where the gene belongs to , the coordinates for the start position, the end position
    and the direction of the gene. 

    Parameters
    ----------
    coordinates : dict 
        This dictionary is the 1st output from the function read_genes(gff_file, essentials_file, gene_names_file)
    chr_lengths_cumsum : dict
        This dictionary is the 2nd output from the function get_sequence_length(bam)
    ref_tid_roman : dict
        This dictionary is the output from 
        ref_tid_roman = {key: value for key, value in zip(ref_romannums, ref_tid)}, where
        ref_romannums = chromosomename_roman_to_arabic()[1] and 
        ref_tid = get_chromosome_names(bam)
        

    Returns
    -------
     dict
        A dictionary that for every gene , there is the chromosome number information of where the gene belongs to , the coordinates for the start position, the end position
    and the direction of the gene. 
    """  

    for key in coordinates:
        gene_chrom = ref_tid_roman.get(coordinates.get(key)[0])
        coordinates[key][1] = coordinates.get(key)[1] + chr_lengths_cumsum.get(gene_chrom)
        coordinates[key][2] = coordinates.get(key)[2] + chr_lengths_cumsum.get(gene_chrom)
    
    return coordinates




from .get_insertions_and_reads import get_insertions_and_reads
from .get_reads import get_reads
from .concatenate_chromosomes import (
    add_chromosome_length,
    add_chromosome_length_inserts,
)
from .find_chromosome_reads import find_chromosome_reads
from .correct_read_position import correct_read_position


__all__ = [
    "get_insertions_and_reads",
    "get_reads",
    "add_chromosome_length",
    "add_chromosome_length_inserts",
    "find_chromosome_reads",
    "correct_read_position"
]



import os


def verify_data_files(path: str = "", data_files: dict = {}):
    """
    Verify presence of the following essential data files
    - bam
    - gff3
    - essential genes
    - gene names

    Parameters
    ----------
    path : str
        Folder containing all data files.
    data_files : dict
        Dictionary containing filetype and filename as key:value

    TODO: put data_files in a separate config file

    """

    # Verify existence data folder
    assert os.path.isdir(path), f"{path} was not found."

    if not data_files:
        data_files = {
            # "bam": "WT_merged-techrep-a_techrep-b_trimmed.sorted.bam",
            "gff3": "Saccharomyces_cerevisiae.R64-1-1.99.gff3",
            "essential_genes": "Cerevisiae_AllEssentialGenes_List.txt",
            "gene_names": "Yeast_Protein_Names.txt",
        }

    # Check presence data files
    for filetype, file_name in data_files.items():
        file_path = os.path.join(path, file_name)
        assert os.path.isfile(file_path), f"{filetype} not found at {file_path}"



import os


def save_as_bed(file, tncoordinates_array, ref_tid, readnumb_array):
    """This function writes in a .bed file located in the same folder as the bamfile, the information abot how many transposons
    were found in each genomic location. 

    Parameters
    ----------
    file : str
        Path with the filename extension included(e.g "data_file/bedfile_name.bed") describing where do you want to store the results.
        By default it will be stored in the same location as the bamfile, with the same basename. 
        Example, if the bamfile path is data_file/data_1.bam then the bed file will be data_file/data_1.bam.bed 
    tncoordinates_array : numpy array 
        Second Output from the get_reads.py function : _,tncoordinates_array,_=get_reads(bam)
    ref_tid : dict 
        Output from the get_chromosome_names function : ref_tid = get_chromosome_names(bam)
    readnumb_array : numpy array 
        1st Output from the get_reads.py function : readnumb_array,_,_=get_reads(bam)
    """

    with open(file, "w") as f:

        base = os.path.basename(file)
        filename = os.path.splitext(base)[0]
        f.write("track name=" + filename + " useScore=1\n")

        coordinates_counter = 0
        for tn in tncoordinates_array:
            refname = [key for key, val in ref_tid.items() if val == tn[0] - 1][0]
            if refname == "Mito":
                refname = "M"
            f.write(
                "chr"
                + refname
                + " "
                + str(tn[1])
                + " "
                + str(tn[1] + 1)
                + " . "
                + str(100 + readnumb_array[coordinates_counter] * 20)
                + "\n"
            )
            coordinates_counter += 1



def gene_position(gff_file, get_dict=True):
    """Get the start and end position of each gene and determine their respective length.
    Input is a .gff file downloaded from https://www.ensembl.org/Saccharomyces_cerevisiae/Info/Index
    Output is a dictionary that includes all gene names as keys. The values are lists with four inputs.
    The first is the chromosome number the gene belong to, the second is the start position, the third is the end position of the gene in terms of basepairs, the fourth is the reading orientation of the gene.
    The reading orientation is indicated with a '+' (forward reading) or '-' (reverse reading).
    The get_dict by default sets that the output should be given as a dictionary with keys the different genes and the values a list of the different parameters.
    When the get_dict is set to False, the code returns all the values as individual lists.

    Parameters
    ----------
    gff_file : str
        The file path of  a .gff file downloaded from https://www.ensembl.org/Saccharomyces_cerevisiae/Info/Index
    get_dict : bool, optional
        When the get_dict is set to False, the code returns all the values as individual lists, by default True

    Returns
    -------
    dict
        Output is a dictionary that includes all gene names as keys. The values are lists with four inputs.
        The first is the chromosome number the gene belong to, the second is the start position, the third is the end position of the gene in terms of basepairs, the fourth is the reading orientation of the gene.
        The reading orientation is indicated with a '+' (forward reading) or '-' (reverse reading).
    """    

    if get_dict == True:
        gene_pos_dict = {}
        with open(gff_file) as f:
            for line in f:
                line_list = line.split("\t")
                if len(line_list) > 2:
                    if line_list[2] == "gene":
                        gene_chr = line_list[0]
                        gene_start = line_list[3]
                        gene_end = line_list[4]
                        gene_orien = line_list[6]
                        gene_position = [
                            gene_chr,
                            int(gene_start),
                            int(gene_end),
                            gene_orien,
                        ]

                        gene_name_string = line_list[8].split(";")[0]
                        gene_name = gene_name_string.split(":")[1]

                        gene_pos_dict[gene_name] = gene_position

        return gene_pos_dict

    else:
        gene_chr = []
        gene_start = []
        gene_end = []
        gene_orien = []
        gene_name = []
        with open(gff_file) as f:
            for line in f:
                line_list = line.split("\t")
                if len(line_list) > 2:
                    if line_list[2] == "gene":
                        gene_chr.append(line_list[0])
                        gene_start.append(int(line_list[3]))
                        gene_end.append(int(line_list[4]))
                        gene_orien.append(line_list[6])

                        gene_name_string = line_list[8].split(";")[0]
                        gene_name.append(gene_name_string.split(":")[1])

        return (gene_name, gene_chr, gene_start, gene_end, gene_orien)



from .verify_data_files import verify_data_files
from .load_default_files import load_default_files,load_sgd_tab
from .read_genes import read_genes


__all__ = [
    "load_default_files",
    "read_genes",
    "load_sgd_tab"
]




import os
import numpy as np
import matplotlib.pyplot as plt


from transposonmapper.properties.get_chromosome_position import chromosome_position
from transposonmapper.properties.get_gene_position import  gene_position
from transposonmapper.processing.chromosome_names_in_files import chromosome_name_bedfile
from transposonmapper.processing.essential_genes_names import list_known_essentials

from transposonmapper.importing import load_default_files

from transposonmapper.plotting import profile_genome_plot

from transposonmapper.processing.profileplot_genome_helpers import (summed_chr,
length_genome,middle_chrom_pos,counts_genome,binned_list)


def profile_genome(bed_file=None, variable="transposons", bar_width=None, savefig=False,showfig=False):
    """Created on Thu Mar 18 13:05:39 2021

    @author: gregoryvanbeek
    This function creates a bar plot along the entire genome.
    The height of each bar represents the number of transposons or reads at the genomic position indicated on the x-axis.
   
    The bar_width determines how many basepairs are put in one bin. Little basepairs per bin may be slow. 
    Too many basepairs in one bin and possible low transposon areas might be obscured.
    

    Parameters
    ----------
    bed_file : str, optional
        The file path to the location of the bed file in your filesystem, by default None
    variable : str, optional
        The variable for plotting throughput the genome, by default "transposons"
    bar_width : int, optional
        The width for the histogram of the plot, by default None , which means internally the length of the genome over 1000 
    savefig : bool, optional
        Save the figure if True, by default False
    showfig : bool, optional
        Show the figure if True, by default False

    Returns
    -------
    list
        All insertion sites
    list
        Binned insertion sites according the width 
    """




    # If necessary, load default files
    gff_file, essential_file, gene_name_file = load_default_files(
        gff_file=None, essentials_file=None, gene_names_file=None
    )

    # Verify presence of files
    data_files = {
        "gff3": gff_file,
        "essentials": essential_file,
        "gene_names": gene_name_file,
    }

    for filetype, file_path in data_files.items():
        assert file_path, f"{filetype} not found at {file_path}"


    chrom_list = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX', 'X', 'XI', 'XII', 'XIII', 'XIV', 'XV', 'XVI']
    
    chr_length_dict, chr_start_pos_dict, chr_end_pos_dict = chromosome_position(gff_file)
    
   
    summed_chr_length_dict=summed_chr(chr_length_dict)
    
       
    l_genome=length_genome(chr_length_dict)
    
    if bar_width == None:
        bar_width = l_genome/1000
    
    print('Genome length: ', l_genome)
    
       
    middle_chr_position=middle_chrom_pos(chr_length_dict)

    gene_pos_dict = gene_position(gff_file)
    
    genes_currentchrom_pos_list = [k for k, v in gene_pos_dict.items()]
    
    genes_essential_list = list_known_essentials(essential_file)


    allcounts_list=counts_genome(variable,bed_file,gff_file)

    allcounts_binnedlist=binned_list(allcounts_list,bar_width)


    if bar_width == (l_genome/1000):
        allinsertionsites_list = np.linspace(0,l_genome,int(l_genome/bar_width+1))
    else:
        allinsertionsites_list = np.linspace(0,l_genome,int(l_genome/bar_width+2))


    ##########Ploting##############
    
    profile_genome_plot(bar_width,l_genome,allinsertionsites_list,allcounts_binnedlist,summed_chr_length_dict,
                         middle_chr_position,chrom_list,variable,genes_currentchrom_pos_list,gene_pos_dict)
    

    # saving the plot 
    if savefig == True and variable == "transposons":
        savepath = os.path.splitext(bed_file)
        print('saving figure at %s' % savepath[0]+'_transposonplot_genome.png')
        plt.savefig(savepath[0]+'_transposonplot_genome.png', dpi=400)
        plt.close()
    elif savefig == True and variable == "reads":
        savepath = os.path.splitext(bed_file)
        print('saving figure at %s' % savepath[0]+'_readplot_genome.png')
        plt.savefig(savepath[0]+'_readplot_genome.png', dpi=400)
        plt.close()
    if showfig==True:
        plt.show()
        
    return allinsertionsites_list,allcounts_binnedlist






import os
def chromosome_name_bedfile(bed_file):
    '''This function returns some properties of the chromosomes in a bed file.
    Input can be either of two options:
        The full path to a bed file after which the program opens the bed file, or
        a list of the lines in the bed file. The latter requires to read the bed file before calling this function and input all lines in the bed file as a list. The function than does not open the bed file again.
    Returns three dictionaries (in this order):
        The first indicates the names of the chromosomes as used in the bed file (keys are roman numerals 1 to 16 and the values are the names used in the bed file).
        The second is the start line in the bed file of each chromosome (keys are the roman numerals of the chromosome names and the values are the start lines in the bed file of the chromosome).
        The third is the end line in the bed file of each chromosome (keys are the roman numerals of the chromosome names and the values are the start lines in the bed file of the chromosome)

    CHANGE LINE 60 AND 71 TO AUTOMATICALLY RECOGNIZE THE MITOCHONDRIAL DNA NAME
    '''
    
    
    if type(bed_file) is str:
        
        assert os.path.isfile(bed_file), "ERROR: bed file not found"
        with open(bed_file) as f:
            lines = f.readlines()
    elif type(bed_file) is list:
        lines = bed_file




    num_arabic = [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16]
    num_roman = ['I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI']
    roman_to_arabic_dict = {}
    index_counter = 0
    for rom in num_roman:
        roman_to_arabic_dict[rom] = num_arabic[index_counter]
        index_counter += 1

    chromosome_romannames_list = []
    for roman in roman_to_arabic_dict:
        chromosome_romannames_list.append(roman)




    chrom_names_dict = {}
    chrom_start_line_dict = {}
    chrom_end_line_dict = {}

    chrom_name_in_bed = ''
    chr_counter = 0
    line_counter = 0
    stop_loop = False
    while stop_loop is False:
        line = lines[line_counter]
        chrom_name_current = line.split(' ')[0].replace('chr','')
        if not chrom_name_current.startswith('track') and not chrom_name_current.startswith('M') and not chrom_name_current.startswith('ref|NC_001224|') and chr_counter < 16: #SKIP HEADER AND MITOCHRONDRIAL CHROMOSOMES #and not chrom_name_current.startswith('M')
            if chrom_name_current != chrom_name_in_bed:
                chrom_names_dict[chromosome_romannames_list[chr_counter]] = chrom_name_current
                chrom_name_in_bed = chrom_name_current
#                print('Chromosome ',chromosome_romannames_list[chr_counter], 'is ',chrom_name_current)
                
                chrom_start_line_dict[chromosome_romannames_list[chr_counter]] = line_counter #GET START INDEX IN THE BED FILE OF THE CURENT CHROMOSOME
                if chr_counter != 0:
                    chrom_end_line_dict[chromosome_romannames_list[chr_counter-1]] = line_counter-1 #GET THE END INDEX IN THE BED OF THE PREVIOUS CHROMOSOME (SKIP FOR THE FIRST CHROMOSOME)

                chr_counter += 1

        elif chrom_name_current.startswith('M') or chrom_name_current.startswith('ref|NC_001224|'):
            chrom_end_line_dict[chromosome_romannames_list[-1]] = line_counter-1 #GET THE END INDEX IN THE BED FILE FOR THE FINAL CHROMOSOME
            stop_loop = True
                
        line_counter += 1
        
    return(chrom_names_dict, chrom_start_line_dict, chrom_end_line_dict)

#%%
def chromosome_name_wigfile(wig_file):

    if type(wig_file) is str:
      
        assert os.path.isfile(wig_file), "ERROR: wig file not found"
        with open(wig_file) as f:
            lines = f.readlines()
    elif type(wig_file) is list:
        lines = wig_file




    num_arabic = [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17]
    num_roman = ['I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','M']
    roman_to_arabic_dict = {}
    index_counter = 0
    for rom in num_roman:
        roman_to_arabic_dict[rom] = num_arabic[index_counter]
        index_counter += 1

    chromosome_romannames_list = []
    for roman in roman_to_arabic_dict:
        chromosome_romannames_list.append(roman)





    chrom_names_dict = {}
    chrom_start_line_dict = {}
    chrom_end_line_dict = {}

    chr_counter = 0
    line_counter = 0
    chrom_line = 'variablestep'
    for line in lines:
        line.strip('\n')
        line_split = line.split(' ')
        if line_split[0].lower() == chrom_line and chr_counter < 17:
            chromosome_name = line_split[1].replace('chrom=chr','').strip('\n')
            chrom_names_dict[chromosome_romannames_list[chr_counter]] = chromosome_name
#            print('Chromosome ',chromosome_romannames_list[chr_counter], 'is ',chromosome_name)
            
            chrom_start_line_dict[chromosome_romannames_list[chr_counter]] = line_counter+1 #GET START INDEX IN THE BED FILE OF THE CURENT CHROMOSOME
            if chr_counter != 0:
                chrom_end_line_dict[chromosome_romannames_list[chr_counter-1]] = line_counter #GET THE END INDEX IN THE BED OF THE PREVIOUS CHROMOSOME (SKIP FOR THE FIRST CHROMOSOME)

            chr_counter += 1
        line_counter += 1

    chrom_end_line_dict[chromosome_romannames_list[chr_counter-1]] = len(lines)


    return(chrom_names_dict, chrom_start_line_dict, chrom_end_line_dict)

# #%%
# if __name__ == '__main__':
#     chrom_names_dict, chrom_start_line_dict, chrom_end_line_dict = chromosome_name_bedfile(bed_file=r"")
#     # chrom_names_dict, chrom_start_line_dict, chrom_end_line_dict = chromosome_name_wigfile(wig_file=r"")


import os
from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname("__file__"))

version = {}
with open(os.path.join(here, "transposonmapper", "__version__.py")) as f:
    exec(f.read(), version)


with open("README.rst") as readme_file:
    readme = readme_file.read()

setup(
    name="transposonmapper",
    version=version["__version__"],
    description="A libray for processing sequencing data for SAturated Transposon Analysis in Yeast (SATAY)",
    long_description=readme,
    long_description_content_type='text/x-rst',
    url="https://github.com/SATAY-LL/Transposonmapper",
    author="Liedewij Laan Lab",
    author_email="L.M.InigoDeLaCruz@tudelft.nl",
    license="Apache Software License 2.0",
    packages=find_packages(exclude=["*tests*"]),
    package_data={"transposonmapper": ["data_files/*"]},
    key_words=["transposon-mapping", "Saccharomyces Cerevisiae",],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Unix Shell",
    ],
    test_suite="tests",
    install_requires=["numpy", "matplotlib", "scipy","pandas"],
    extras_require={
        "dev": [
            "bump2version",
            "mock",
            "pysam",
            "pytest",
            "pytest-cov",
            "jupyter-book>=0.7.0",
            "sphinx-click",
            "sphinx-tabs",
            "sphinxext-rediraffe",
            "sphinx_inline_tabs",
            "ghp-import"

        ],
        "linux":[
            "pysam"

        ]

    },
)



import os, sys

import matplotlib.pyplot as plt
import numpy as np


from transposonmapper.statistics.volcano_helpers import apply_stats,  info_from_datasets, make_datafile

def volcano(path_a, filelist_a, path_b, filelist_b,fold_change_interval,p_value_interval,variable='read_per_gene',
significance_threshold=0.01, normalize=True, trackgene_list=[], figure_title="",savefigure=False):
    """This script creates a volcanoplot to show the significance of fold change between two datasets.
    It is based on this website:
        - https://towardsdatascience.com/inferential-statistics-series-t-test-using-numpy-2718f8f9bf2f
        - https://www.statisticshowto.com/independent-samples-t-test/

    Code for showing gene name when hovering over datapoint is based on:
        - https://stackoverflow.com/questions/7908636/possible-to-make-labels-appear-when-hovering-over-a-point-in-matplotlib

    T-test is measuring the number of standard deviations our measured mean is from the baseline mean, while taking into
    account that the standard deviation of the mean can change as we get more data
    This creates a volcano plot that shows the fold change between two libraries and the corresponding p-values.
    
    The fold change is determined by the mean of dataset b (experimental set) divided by the mean of dataset a (reference set).
    The datasets can be of different length.
    P-value is determined based on the student t-test (scipy.stats.ttest_ind).

    NOTE:
        The fold change is determined by the ratio between the reference and the experimental dataset.
        When one of the datasets is 0, this is false results for the fold change.
        To prevent this, the genes with 0 insertions are set to have 5 insertions, and the genes with 0 reads are set to have 25 reads.
        These values are determined in dicussion with the Kornmann lab.

    - Created on Tue Feb 16 14:06:48 2021
    - @author: gregoryvanbeek

 
    Parameters
    ----------
    path_a : str
        paths to location of the datafiles for library a 
    filelist_a : str
        list of the names of the datafiles for library a  located in path_a. 
        The type of file here is the pergene.txt file , which is one of the outputs from  the transposonmapper function. 
        The format of the pergene file should be TAB separated and NOT COMMA separated. if you have
        it as comma separated you can convert to tab separated using the command line with 
        this command: `cat oldfile.txt | tr '[,]' '[\t]' > newfile.txt`
    path_b : str
        paths to location of the datafiles for library b
    filelist_b : str
        list of the names of the datafiles for  library b located in path_b 
        The type of file here is the pergene.txt file , which is one of the outputs from  the transposonmapper function. 
        The format of the pergene file should be TAB separated and NOT COMMA separated. if you have
        it as comma separated you can convert to tab separated using the command line with 
        this command: `cat oldfile.txt | tr '[,]' '[\t]' > newfile.txt`

    fold_change_interval : numpy.array

            a vector of two elements where the first element(positive value) set the upper threshold for the positive
            values and the second element(negative value) set the lower threshold for the negative values.


    p_value_interval : numpy.array
            a vector of two elements where the first element set the upper threshold for the positive
           values and the second element set the lower threshold for the negative values.
    variable : str, optional
        tn_per_gene, read_per_gene or Nreadsperinsrt , by default 'read_per_gene'
    significance_threshold : float, optional
        Threshold value above which the fold change is regarded significant, only for plotting, by default 0.01
    normalize : bool, optional
        Whether to normalize variable. If set to True, each gene is normalized based on the total count in each dataset (i.e. each file in filelist_)
        , by default True
    trackgene_list : list, optional
        Enter a list of gene name(s) which will be highlighted in the plot (e.g. ['cdc42', 'nrp1']), by default []
    figure_title : str, optional
        The title of the figure if not empty, by default ""
    savefigure : bool, optional


    Returns
    -------
    dataframe

        A dataframe containing:
        
            - gene_names
            - fold change
            - t statistic
            - p value
            - whether p value is above threshold
    figure
        - volcanoplot with the log2 fold change between the two libraries and the -log10 p-value.

    """

### Making the whole datafile name 

    datafiles_list_a,datafiles_list_b=make_datafile(path_a,filelist_a,path_b,filelist_b)


### Extract information from datasets

    variable_a_array,variable_b_array,volcano_df,tnread_gene_a,_=info_from_datasets(datafiles_list_a,datafiles_list_b,variable,normalize)

### APPLY stats.ttest_ind(A,B)
    volcano_df=apply_stats(variable_a_array,variable_b_array,significance_threshold,volcano_df)

   


### Volcanoplot
    print('Plotting: %s' % variable)

    fig = plt.figure(figsize=(19.0,9.0))#(27.0,3))
    grid = plt.GridSpec(1, 1, wspace=0.0, hspace=0.0)
    ax = plt.subplot(grid[0,0])

    colors = {False:'gray', True:'red'} # based on p-value significance 
    sc = ax.scatter(x=volcano_df['fold_change'], y=volcano_df['p_value'], alpha=0.4, 
    s=100, c=volcano_df['significance'].apply(lambda x:colors[x]),label=variable)
    ax.grid(True, which='major', axis='both', alpha=0.4)
    ax.set_xlabel('Log2 FC',fontsize=16)
    ax.set_ylabel('-*Log10 p-value',fontsize=16)
    ax.tick_params(axis='both', which='major', labelsize=14)
    ax.legend(loc='best',fontsize=16)
    if not figure_title == "":
        ax.set_title(figure_title,fontsize=20)
    else:
        ax.set_title(variable,fontsize=20)
    # ax.scatter(x=[],y=[],marker='.',color='black', label='p-value > {}'.format(significance_threshold)) #set empty scatterplot for legend
    # ax.scatter(x=[],y=[],marker='.',color='red', label='p-value < {}'.format(significance_threshold)) #set empty scatterplot for legend
    # ax.legend()
    if not trackgene_list == []:
        genenames_array = volcano_df['gene_names'].to_numpy()
        for trackgene in trackgene_list:
            trackgene = trackgene.upper()
            if trackgene in genenames_array:
                trackgene_index = tnread_gene_a.loc[tnread_gene_a['gene_names'] == trackgene].index[0]
                trackgene_annot = ax.annotate(volcano_df.iloc[trackgene_index,:]['gene_names'], (volcano_df.iloc[trackgene_index,:]['fold_change'], volcano_df.iloc[trackgene_index,:]['p_value']),
                            size=12, c='green', bbox=dict(boxstyle="round", fc="w"))
                trackgene_annot.get_bbox_patch().set_alpha(0.6)
            else:
                print('WARNING: %s not found' % trackgene)
        


    names = volcano_df['gene_names'].to_numpy()
    annot = ax.annotate("", xy=(0,0), xytext=(20,20),textcoords="offset points",
                        bbox=dict(boxstyle="round", fc="w"),
                        arrowprops=dict(arrowstyle="->"))
    annot.set_visible(False)

    ## Annotate based on fold change and p values 
    if not fold_change_interval == [] or not p_value_interval == []:
        target=volcano_df[(volcano_df["fold_change"]>fold_change_interval[0]) & (volcano_df["p_value"]>p_value_interval[0])]["gene_names"]
        target_left=volcano_df[(volcano_df["fold_change"]<fold_change_interval[1]) & (volcano_df["p_value"]>p_value_interval[1])]["gene_names"]

        if len(target)!=0:
            for i in np.arange(0,len(target)):

                index=np.where(volcano_df==target.iloc[i])[0][0]

            

                trackgene_annot = ax.annotate(volcano_df.iloc[index,:]['gene_names'], (volcano_df.iloc[index,:]['fold_change'],
                                            volcano_df.iloc[index,:]['p_value']),
                                            size=12, c='green', bbox=dict(boxstyle="round", fc="w"))
                trackgene_annot.get_bbox_patch().set_alpha(0.6)

        if len(target_left)!=0:
            for i in np.arange(0,len(target_left)):

                index=np.where(volcano_df==target_left.iloc[i])[0][0]

            

                trackgene_annot = ax.annotate(volcano_df.iloc[index,:]['gene_names'], (volcano_df.iloc[index,:]['fold_change'],
                                            volcano_df.iloc[index,:]['p_value']),
                                            size=12, c='green', bbox=dict(boxstyle="round", fc="w"))
                trackgene_annot.get_bbox_patch().set_alpha(0.6)

    

    
    def update_annot(ind):
    
        pos = sc.get_offsets()[ind["ind"][0]]
        annot.xy = pos
        # text = "{}, {}".format(" ".join(list(map(str,ind["ind"]))), 
        #                         " ".join([names[n] for n in ind["ind"]]))
        text = "{}".format(" ".join([names[n] for n in ind["ind"]]))
        annot.set_text(text)
        # annot.get_bbox_patch().set_facecolor(cmap(norm(c[ind["ind"][0]])))
        # annot.get_bbox_patch().set_alpha(0.4)


    def hover(event):
        vis = annot.get_visible()
        if event.inaxes == ax:
            cont, ind = sc.contains(event)
            if cont:
                update_annot(ind)
                annot.set_visible(True)
                fig.canvas.draw_idle()
            else:
                if vis:
                    annot.set_visible(False)
                    fig.canvas.draw_idle()
                    
    fig.canvas.mpl_connect("motion_notify_event", hover)

    if savefigure:
        fig.savefig("volcano_"+filelist_a[0].split("_")[0]+"vs"+filelist_b[0].split("_")[0]+".png", dpi=400)


## return function
    return(volcano_df)


